select(2, ...) 'aux.core.scan'

local aux = require 'aux'
local info = require 'aux.util.info'
local history = require 'aux.core.history'

local DEFAULT_PAGE_SIZE = 50
local TIMEOUT = 30

function aux.handle.CLOSE()
	abort()
end

function auctions(type)
    local auctions = {}
    for i = 1, GetNumAuctionItems(type) do
        local auction_record = info.auction(i, type)
        if auction_record then
            auctions[i] = auction_record
        end
    end
    return pairs(auctions)
end

function M.owner_auctions()
    return auctions('owner')
end

function M.bidder_auctions()
    return auctions('bidder')
end

do
	function M.start(params)
		if state then
			abort(state.id)
		end
		do (params.on_scan_start or pass)() end
        local thread_id = aux.coro_thread(function()
            state = {
                id = aux.coro_id(),
                params = params,
            }
            scan()
        end)
        return thread_id
	end

	function M.abort()
        if state then
            aux.coro_kill(state.id)
            aux.kill_listener(state.listener_id)
            local on_abort = state.params.on_abort
            state = nil
            do (on_abort or pass)() end
        end
	end

	function M.stop()
		state.stopped = true
	end

	function complete()
		local on_complete = state.params.on_complete
        state = nil
		do (on_complete or pass)() end
	end
end

function get_query()
    local queries
    if state.params.get_all then
        queries =  {{blizzard_query = {}}}
    else
        queries = state.params.queries
    end
	return queries[state.query_index]
end

function total_pages(total_auctions)
    return page_size() == 0 and 0 or ceil(total_auctions / page_size())
end

function last_page(total_auctions)
    local last_page = max(total_pages(total_auctions) - 1, 0)
    local last_page_limit = get_query().blizzard_query.last_page or last_page
    return min(last_page_limit, last_page)
end

function page_size()
    if not state.params.get_all then
        return GetNumAuctionItems('list')
    else
        return DEFAULT_PAGE_SIZE
    end
end

function scan()
    state.query_index = 1
	while get_query() and not state.stopped do
		do (state.params.on_start_query or pass)(state.query_index) end
		if get_query().blizzard_query then
            state.page = get_query().blizzard_query.first_page or 0
            while state.page <= (get_query().blizzard_query.last_page or math.huge) do
				submit_query()
                wait_for_results()
                state.page = state.page + 1
                if state.page > last_page(state.total_auctions) then
                    break
                end
            end
		else
			state.page = nil
			scan_page()
        end
        state.query_index = state.query_index + 1
    end
	complete()
end

function submit_query()
    if state.stopped then
        return
    end

    while not CanSendAuctionQuery() do
        aux.coro_wait()
    end

    SortAuctionClearSort('list')
    SortAuctionItems('list', 'duration')
    SortAuctionItems('list', 'duration')

    state.last_list_query = GetTime()
    local blizzard_query = get_query().blizzard_query or {}
    local category_filter
    if blizzard_query.class and blizzard_query.subclass and blizzard_query.slot then
        category_filter = AuctionCategories[blizzard_query.class].subCategories[blizzard_query.subclass].subCategories[blizzard_query.slot].filters
    elseif blizzard_query.class and blizzard_query.subclass then
        category_filter = AuctionCategories[blizzard_query.class].subCategories[blizzard_query.subclass].filters
    elseif blizzard_query.class then
        category_filter = AuctionCategories[blizzard_query.class].filters
    else
        -- not filtering by category, leave nil for all
    end
    QueryAuctionItems(
        blizzard_query.name,
        blizzard_query.min_level,
        blizzard_query.max_level,
        state.page,
        blizzard_query.usable,
        blizzard_query.quality,
        state.params.get_all,
        blizzard_query.class ~= 1 and blizzard_query.class ~= 2 and blizzard_query.exact, -- Doesn't work for suffix items
        category_filter
    )
end

function scan_page(results)
    for i = 1, page_size() do
        if i % 1000 == 0 then -- Throttling for getAll scan
            local t0 = GetTime()
            while GetTime() < t0 + .1 do
                aux.coro_wait()
            end
        end
        local auction_info = results and results[i] or info.auction(i)
        if auction_info and (auction_info.owner or state.params.ignore_owner or aux.account_data.ignore_owner) then -- TODO
            auction_info.index = i
            auction_info.page = state.page
            auction_info.blizzard_query = get_query().blizzard_query

            history.process_auction(auction_info)

            if not get_query().validator or get_query().validator(auction_info) then
                do (state.params.on_auction or pass)(auction_info) end
            end
        end
    end
    do (state.params.on_page_scanned or pass)() end
end

function accept_results(results)
	_,  state.total_auctions = GetNumAuctionItems('list')
	do
		(state.params.on_page_loaded or pass)(
			state.page - (get_query().blizzard_query.first_page or 0) + 1,
			last_page(state.total_auctions) - (get_query().blizzard_query.first_page or 0) + 1,
			total_pages(state.total_auctions) - 1
		)
	end
	scan_page(results)
end

function wait_for_results()
    local scanned_auctions = {}
    local updated, last_update

    state.listener_id = aux.event_listener('AUCTION_ITEM_LIST_UPDATE', function()
        last_update = GetTime()
        updated = true
    end)

    local timeout = aux.later(TIMEOUT, state.last_list_query)

    while true do
        if not last_update and timeout() then
            break
        elseif last_update and GetTime() - last_update > TIMEOUT then
            break
        elseif updated then
            updated = false
            local complete = true
            local count = 0
            for i = 1, GetNumAuctionItems('list') do
                if not scanned_auctions[i] then
                    local auction = info.auction(i, 'list')
                    if auction then
                        scanned_auctions[i] = auction
                    else
                        complete = false
                    end
                    count = count + 1
                    if count % 100 == 0 then
                        local t0 = GetTime()
                        while GetTime() < t0 + .1 do
                            aux.coro_wait()
                        end
                    end
                end
            end
            if complete then
                break
            end
        end
        aux.coro_wait()
    end

    aux.kill_listener(state.listener_id)
    if not last_update and timeout() then
        submit_query()
        wait_for_results()
    else
        accept_results(scanned_auctions)
    end
end